# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hardware parameters of robottime devices."""

from robodyno.components.config.model import Model

ROBOTTIME_PARAMS = {
    'motor': {
        Model.ROBODYNO_PRO_P44: {
            'reduction': -44,
            'available_velocity': 27,
            'available_torque': 13,
            'available_current': 15,
            'torque_constant': 0.7742,
            'with_brake': False,
            'absolution_encoder': False,
            'default_vel_limit': 5.71,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_P12: {
            'reduction': -12.45,
            'available_velocity': 95,
            'available_torque': 4,
            'available_current': 15,
            'torque_constant': 0.2191,
            'with_brake': False,
            'absolution_encoder': False,
            'default_vel_limit': 20.19,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_P44A: {
            'reduction': -44,
            'available_velocity': 27,
            'available_torque': 13,
            'available_current': 15,
            'torque_constant': 0.7742,
            'with_brake': False,
            'absolution_encoder': True,
            'default_vel_limit': 5.71,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_P12A: {
            'reduction': -12.45,
            'available_velocity': 95,
            'available_torque': 4,
            'available_current': 15,
            'torque_constant': 0.2191,
            'with_brake': False,
            'absolution_encoder': True,
            'default_vel_limit': 20.19,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_H50: {
            'reduction': 50,
            'available_velocity': 23.6,
            'available_torque': 15,
            'available_current': 15,
            'torque_constant': 0.88,
            'with_brake': False,
            'absolution_encoder': True,
            'default_vel_limit': 5.03,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_H100: {
            'reduction': 100,
            'available_velocity': 11.8,
            'available_torque': 30,
            'available_current': 15,
            'torque_constant': 1.76,
            'with_brake': False,
            'absolution_encoder': True,
            'default_vel_limit': 2.51,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_B100: {
            'reduction': 100,
            'available_velocity': 11.8,
            'available_torque': 30,
            'available_current': 15,
            'torque_constant': 1.76,
            'with_brake': False,
            'absolution_encoder': True,
            'default_vel_limit': 2.51,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PRO_DIRECT: {
            'reduction': 1,
            'available_velocity': 1180,
            'available_torque': 0.3,
            'available_current': 15,
            'torque_constant': 0.0176,
            'with_brake': False,
            'absolution_encoder': False,
            'default_vel_limit': 251.33,
            'default_current_limit': 13.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.02,
            'default_vel_ki': 0.1,
        },
        Model.ROBODYNO_PLUS_H50: {
            'reduction': 50,
            'available_velocity': 24,
            'available_torque': 48,
            'available_current': 25,
            'torque_constant': 1.7229,
            'with_brake': True,
            'absolution_encoder': True,
            'default_vel_limit': 3.77,
            'default_current_limit': 25.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.10,
            'default_vel_ki': 0.50,
        },
        Model.ROBODYNO_PLUS_H100: {
            'reduction': 100,
            'available_velocity': 12,
            'available_torque': 96,
            'available_current': 25,
            'torque_constant': 3.4458,
            'with_brake': True,
            'absolution_encoder': True,
            'default_vel_limit': 1.88,
            'default_current_limit': 25.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.10,
            'default_vel_ki': 0.50,
        },
        Model.ROBODYNO_PLUS_P12: {
            'reduction': -12.45,
            'available_velocity': 96,
            'available_torque': 12,
            'available_current': 25,
            'torque_constant': 0.429,
            'with_brake': True,
            'absolution_encoder': False,
            'default_vel_limit': 20.19,
            'default_current_limit': 25.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.10,
            'default_vel_ki': 0.50,
        },
        Model.ROBODYNO_PLUS_P12A: {
            'reduction': -12.45,
            'available_velocity': 96,
            'available_torque': 12,
            'available_current': 25,
            'torque_constant': 0.429,
            'with_brake': True,
            'absolution_encoder': True,
            'default_vel_limit': 20.19,
            'default_current_limit': 25.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.10,
            'default_vel_ki': 0.50,
        },
        Model.ROBODYNO_PLUS_DIRECT: {
            'reduction': 1,
            'available_velocity': 1200,
            'available_torque': 0.96,
            'available_current': 25,
            'torque_constant': 0.034458,
            'with_brake': True,
            'absolution_encoder': False,
            'default_vel_limit': 188.50,
            'default_current_limit': 25.0,
            'default_pos_kp': 100.0,
            'default_vel_kp': 0.10,
            'default_vel_ki': 0.50,
        },
        Model.ROBODYNO_NANO_P100: {
            'reduction': 100,
            'available_velocity': 500,  # for develop
            'available_torque': 100,  # for develop
            'available_current': 100,  # for develop
            'torque_constant': 0.25,
            'with_brake': False,
            'absolution_encoder': True,
            'default_vel_limit': 10.0,
            'default_current_limit': 0.5,
            'default_pos_kp': 25.0,
            'default_vel_kp': 200.0,
            'default_vel_ki': 5.0,
        },
    }
}



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a function to generate a table of motor information."""

from time import time
from enum import Enum

from rich.table import Table
from rich.style import Style

from robodyno.components.can_bus.motor import Motor

def error_text(error: Enum) -> str:
    """Returns a rich text object for an error.

    Args:
        error (Enum): The error to get text for.

    Returns:
        (string): The rich text object.
    """
    if error is None:
        return ''
    if error.value:
        return f'[red]{error.name}[/]'
    return f'[green]{error.name}[/]'


def motor_info_table(motor: Motor) -> Table:
    """Returns a table of motor information.

    Args:
        motor (Motor): The motor to get information from.
    Returns:
        (Table): The table of motor information.
    """
    table = Table(
        title=(
            f'[cyan][0x{motor.id:02X}][/] '
            f'[green]{motor.type.name}[/] '
            f'[yellow]{motor.fw_ver}[/]'
        ),
        title_style=Style(italic=False, bold=True)
    )
    table.add_column('Property', style='cyan', justify='left', no_wrap=True)
    table.add_column('Value', style='green', justify='right', no_wrap=True)
    table.add_column(style='yellow', justify='left', no_wrap=True)

    if motor.fw_ver < 1:
        sync_time = motor.sync_time
        t = time()
        timeout = 1
        while motor.sync_time == sync_time and time() - t < timeout:
            pass
        state = motor.state
        error = motor.error
        mode = motor.mode
    else:
        state, error, mode = motor.get_state(timeout = 0.1)
    error_detail = None
    if error['error'] == motor.MotorError.MOTOR_FAILED:
        error_detail = error['motor_err']
    elif error['error'] == motor.MotorError.ENCODER_FAILED:
        error_detail = error['encoder_err']
    elif error['error'] == motor.MotorError.CONTROLLER_FAILED:
        error_detail = error['controller_err']
    table.add_row('State', f'{state.name}', '')
    table.add_row('Error', error_text(error['error']), error_text(error_detail))
    table.add_row('Mode', f'{mode.name}', '')

    voltage = motor.get_voltage(timeout = 0.1)
    table.add_row('Voltage', f'{voltage:.2f}', 'V')
    temperature = motor.get_temperature(timeout = 0.1)
    table.add_row('Temperature', f'{temperature:.2f}', '℃')

    position = motor.get_pos(timeout = 0.1)
    table.add_row('Position', f'{position:.4f}', 'rad')
    if motor.fw_ver >= 1:
        absolute_position = motor.get_abs_pos(timeout = 0.1)
        table.add_row('Absolute Position', f'{absolute_position:.4f}', 'rad')
    velocity = motor.get_vel(timeout = 0.1)
    table.add_row('Velocity', f'{velocity:.4f}', 'rad/s')
    torque = motor.get_torque(timeout = 0.1)
    table.add_row('Torque', f'{torque:.4f}', 'N*m')
    return table



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GPS sensor driver.

This module provides a class to get the GPS data from the GPS sensor which
is based on a GT-U12 module.

Examples:

    >>> from robodyno.components import GpsSensor
    >>> from robodyno.interfaces import CanBus
    >>> can = CanBus()
    >>> gps = GpsSensor(can)
    >>> gps.get_position()
    {'timestamp': datetime.time(0, 0, 0, 0), 'longitude': 0.0, 'latitude': 0.0}
    >>> can.disconnect()
"""

import struct
import datetime
from typing import Optional

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model


class GpsSensor(CanBusDevice):
    """GPS sensor driver.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Firmware version.
        timestamp (float): Timestamp of the last received data.
    """

    _CMD_HEARTBEAT = 0x02
    _CMD_CONFIG_CAN = 0x03
    _CMD_GET_POSITION = 0x04

    def __init__(self, can: CanBus, id_: int = 0x32):
        """Initialize the GPS sensor driver.

        Args:
            can (CanBus): CanBus object.
            id_ (int, optional): Device id. Defaults to 0x32.

        Raises:
            ValueError: If the device type is not GPS sensor.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if self.type is None or self.type != Model.ROBODYNO_GPS_SENSOR:
            raise ValueError(f'GPS sensor init failed, id: {id_}')
        self.timestamp = None
        self._can.subscribe(
            callback=self._on_receive_timestamp,
            device_id=self.id,
            cmd_id=self._CMD_HEARTBEAT,
        )

    def __del__(self):
        try:
            self._can.unsubscribe(
                callback=self._on_receive_timestamp,
                device_id=self.id,
                cmd_id=self._CMD_HEARTBEAT,
            )
        except KeyError:
            pass

    def _on_receive_timestamp(
        self, data: bytes, timestamp: float, device_id: int, cmd_id: int
    ) -> None:
        """Callback function when receiving timestamp data.

        Args:
            data (bytes): Data bytes.
            timestamp (float): Can frame timestamp.
            device_id (int): Device id.
            cmd_id (int): Command id.
        """
        del timestamp, device_id, cmd_id
        hour, minute, second = struct.unpack('<HHf', data)
        self.timestamp = datetime.time(
            hour, minute, int(second), int(second * 1000) % 1000
        )

    def get_position(self, timeout: Optional[float] = None) -> Optional[dict]:
        """Get the position of the GPS sensor.

        Args:
            timeout (float, optional): Timeout. Defaults to None.

        Returns:
            Optional[dict]: Timestamp, longitude and latitude.
        """
        try:
            latitude, longitude = self._can.get(
                self.id,
                self._CMD_GET_POSITION,
                'ff',
                timeout=timeout,
            )
        except TimeoutError:
            return None

        return {
            'timestamp': self.timestamp,
            'longitude': longitude,
            'latitude': latitude,
        }

    def get_position_str(self, timeout: Optional[float] = None) -> Optional[str]:
        """Get the position of the GPS sensor in string format.

        Args:
            timeout (float, optional): Timeout. Defaults to None.

        Returns:
            Optional[str]: Position (timestamp, longitude, latitude) in string format.
        """
        position = self.get_position(timeout)
        if position is None:
            return None
        longitude_str = 'E' if position['longitude'] >= 0 else 'W'
        latitude_str = 'N' if position['latitude'] >= 0 else 'S'
        return (
            f'{self.timestamp} {abs(position.get("longitude"))}°{longitude_str} '
            f'{abs(position.get("latitude"))}°{latitude_str}'
        )

    def config_can_bus(self, new_id: int = None, bitrate: int = 1000000) -> None:
        """Configures the can bus.

        Args:
            new_id (int): New device id.
            bitrate (int): Can bus bitrate.

        Raises:
            ValueError: If the new id is invalid.
        """
        if new_id is None:
            new_id = self.id
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')
        bitrate_id = {
            250000: 0,
            500000: 1,
            1000000: 2,
        }.get(bitrate, 2)
        self._can.send(self.id, self._CMD_CONFIG_CAN, 'HH', new_id, bitrate_id)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a CAN bus monitor tool."""

from queue import Queue

from rich.console import Console
from rich.table import Table
from rich.live import Live

from robodyno.interfaces import CanBus


class CanBusMonitor(object):
    """Monitors all messages on the CAN bus."""

    def __init__(self, can: CanBus, group: bool = False, console: Console = Console()):
        """Initializes the monitor.

        Args:
            can: The CAN bus to monitor.
            group: Whether to group messages by device and command.
            console: The console to use.
        """
        self._can = can
        self._group = group
        self._console = console
        self._msg_queue = Queue()

    def _msg_callback(self, data, us, device_id, command_id) -> None:
        data_list = [f'0x{byte:02X}' for byte in data]
        ts = f'{us // 1000000}.{int((us % 1000000) // 1000):03d}'
        self._msg_queue.put((ts, device_id, command_id, data_list))

    def _generate_table(self, data_dict: dict) -> Table:
        table = Table()
        table.add_column('Timestamp', style='green', justify='center', no_wrap=True)
        table.add_column('ID', style='cyan', justify='center', no_wrap=True)
        table.add_column('Command', style='magenta', justify='center', no_wrap=True)
        table.add_column('Data', style='yellow', justify='left', no_wrap=True)

        for (device_id, command_id), (ts, data_list) in data_dict.items():
            table.add_row(
                ts,
                f'0x{device_id:02X}',
                f'0x{command_id:02X}',
                ' '.join(data_list),
            )
        return table

    def _monitor_grouped(self) -> None:
        data_dict = {}

        with Live(
            self._generate_table(data_dict),
            auto_refresh=False,
            vertical_overflow='visible',
        ) as live:
            while True:
                if self._msg_queue.empty():
                    continue
                ts, device_id, command_id, data_list = self._msg_queue.get()
                data_dict[(device_id, command_id)] = (ts, data_list)
                live.update(self._generate_table(data_dict), refresh=True)

    def _monitor(self) -> None:
        while True:
            if self._msg_queue.empty():
                continue
            ts, device_id, command_id, data_list = self._msg_queue.get()
            self._console.print(
                f'[green not bold][{ts}][/] '
                f'[cyan not bold]0x{device_id:02X}[/] | '
                f'[magenta not bold]0x{command_id:02X}[/] '
                f'[yellow not bold][{" ".join(data_list)}][/]',
            )

    def monitor(self) -> None:
        """Starts monitoring the CAN bus.

        This method blocks until the user presses Ctrl-C.
        """
        self._can.subscribe(self._msg_callback)
        try:
            if self._group:
                self._monitor_grouped()
            else:
                self._monitor()
        except KeyboardInterrupt:
            self._can.unsubscribe(self._msg_callback)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides CAN bus interface for robodyno devices.

Examples:

    >>> from robodyno.interfaces import CanBus
    >>> can_bus = CanBus()
    >>> can_bus.send(0x10, 0x0e, 'fee', 50, 0.02, 0.1)
    >>> can_bus.get(0x10, 0x0d, 'fee')
    (50.0, 0.0200042724609375, 0.0999755859375)
    >>> can_bus.subscribe(callback, 0x10, 0x02)
    >>> can_bus.unsubscribe(callback, 0x10, 0x02)
    >>> can_bus.disconnect()
"""

import sys
from time import time, sleep
from threading import Lock, Thread, Event
from queue import Queue, Empty
from collections import defaultdict
import struct
from typing import Optional, Callable

import can


class CanBus(object):
    """CAN bus interface for robodyno devices.

    This class provides CAN bus interface for robodyno devices. It supports
    sending and receiving data from CAN bus. It also supports subscribing data
    from CAN bus.

    A thread is created to receive data from CAN bus. When a device ID and a
    command ID is subscribed, the callback function will be called when the
    corresponding data is received.
    """

    def __init__(
        self,
        bitrate: int = 1000000,
        channel: str = 'can0',
        connect: bool = True,
    ):
        """Initialize CAN bus interface with given channel and bitrate.

        Args:
            bitrate (int): CAN bus bitrate, e.g. 1000000.
            channel (str): CAN bus channel name, e.g. 'can0'.
            connect (bool): Whether to connect to CAN bus.

        Raises:
            NotImplementedError: If the platform is not supported.
        """
        self._channel = channel
        self._bitrate = bitrate
        if sys.platform == 'win32':
            self._bus_type = 'candle'
        elif sys.platform.startswith('linux'):
            self._bus_type = 'socketcan'
        else:
            raise NotImplementedError(f'Unsupported platform: {sys.platform}')
        self._bus = None
        self._recv_callbacks = defaultdict(lambda: defaultdict(set))
        self._recv_lock = Lock()
        self._send_lock = Lock()
        self._get_lock = Lock()

        self._recv_event = Event()
        self._listening_event = Event()
        self._rx_queue = Queue()

        if connect:
            self.connect()
            sleep(0.1)

    def connect(self) -> None:
        """Connect to CAN bus."""
        self._bus = can.interface.Bus(
            channel=self._channel, bustype=self._bus_type, bitrate=self._bitrate
        )
        self._recv_thread = Thread(target=self._recv_loop, daemon=True)
        self._recv_event.set()
        self._recv_thread.start()

    def disconnect(self) -> None:
        """Disconnect from CAN bus."""
        if self._recv_thread is not None:
            self._recv_event.clear()
            self._recv_thread.join()
            self._recv_thread = None
        if self._bus is not None:
            self._bus.shutdown()
            self._bus = None

    def _combine_ids(self, device_id: int, cmd_id: int) -> int:
        """Combine device ID and command ID to arbitration ID.

        Args:
            device_id (int): Device ID.
            cmd_id (int): Command ID.

        Returns:
            (int): Arbitration ID.
        """
        return (device_id << 5) + cmd_id

    def _split_ids(self, arbitration_id: int) -> tuple:
        """Split arbitration ID to device ID and command ID.

        Args:
            arbitration_id (int): Combined ID.

        Returns:
            (tuple): Device ID and command ID.
        """
        device_id = arbitration_id >> 5
        cmd_id = arbitration_id & 0x1F
        return device_id, cmd_id

    def send(self, device_id: int, cmd_id: int, fmt: str, *args) -> None:
        """Send data to CAN bus.

        Args:
            device_id (int): Device ID.
            cmd_id (int): Command ID.
            fmt (str): Format string for packing data. See [struct module](
                https://docs.python.org/3/library/struct.html#format-characters)
                for details.
            *args (): Data to be sent.
        """
        with self._send_lock:
            arbitration_id = self._combine_ids(device_id, cmd_id)
            data = struct.pack('<' + fmt, *args)
            msg = can.Message(
                arbitration_id=arbitration_id,
                is_extended_id=False,
                data=data,
            )
            self._bus.send(msg)

    def _send_remote(self, device_id: int, cmd_id: int) -> None:
        """Send remote frame to CAN bus.

        Args:
            device_id (int): Device ID.
            cmd_id (int): Command ID.
        """
        with self._send_lock:
            arbitration_id = self._combine_ids(device_id, cmd_id)
            msg = can.Message(
                arbitration_id=arbitration_id,
                is_extended_id=False,
                is_remote_frame=True,
            )
            self._bus.send(msg)

    def get(
        self, device_id: int, cmd_id: int, fmt: str, timeout: Optional[float] = None
    ) -> tuple:
        """Get data from CAN bus.

        Args:
            device_id (int): Device ID.
            cmd_id (int): Command ID.
            fmt (str): Format string for unpacking data. See [struct module](
                https://docs.python.org/3/library/struct.html#format-characters)
                for details.
            timeout (float): Timeout for receiving data.

        Returns:
            (tuple): Received data.

        Raises:
            TimeoutError: If timeout.
        """
        with self._get_lock:
            start_time = time()
            arbitration_id = self._combine_ids(device_id, cmd_id)
            with self._rx_queue.mutex:
                self._rx_queue.queue.clear()
            self._listening_event.set()
            self._send_remote(device_id, cmd_id)
            while True:
                if timeout is not None and time() - start_time > timeout:
                    self._listening_event.clear()
                    raise TimeoutError('Timeout when receiving data.')
                try:
                    msg = self._rx_queue.get(timeout=0.001)
                except Empty:
                    continue
                if msg.arbitration_id == arbitration_id:
                    self._listening_event.clear()
                    return struct.unpack('<' + fmt, msg.data)

    def subscribe(
        self, callback: Callable, device_id: int = -1, cmd_id: int = -1
    ) -> None:
        """Subscribe data from CAN bus.

        Args:
            callback (func): Callback function.

                The function should accept four arguments,
                e.g. `callback(data, timestamp, device_id, cmd_id)`.

            device_id (int): Device ID. -1 for all devices.
            cmd_id (int): Command ID. -1 for all commands.
        """
        with self._recv_lock:
            self._recv_callbacks[device_id][cmd_id].add(callback)

    def unsubscribe(
        self, callback: Callable, device_id: int = -1, cmd_id: int = -1
    ) -> None:
        """Unsubscribe data from CAN bus.

        Args:
            callback (func): Callback function.
            device_id (int): Device ID. -1 for all devices.
            cmd_id (int): Command ID. -1 for all commands.
        """
        with self._recv_lock:
            self._recv_callbacks[device_id][cmd_id].remove(callback)

    def _notify(self, msg: can.Message) -> None:
        """Notify received data."""
        if self._listening_event.is_set():
            self._rx_queue.put(msg)
        device_id, cmd_id = self._split_ids(msg.arbitration_id)
        for callback in self._recv_callbacks[device_id][cmd_id]:
            callback(msg.data, msg.timestamp, device_id, cmd_id)
        for callback in self._recv_callbacks[device_id][-1]:
            callback(msg.data, msg.timestamp, device_id, cmd_id)
        for callback in self._recv_callbacks[-1][cmd_id]:
            callback(msg.data, msg.timestamp, device_id, cmd_id)
        for callback in self._recv_callbacks[-1][-1]:
            callback(msg.data, msg.timestamp, device_id, cmd_id)

    def _recv_loop(self) -> None:
        """Receive data from CAN bus."""
        while self._recv_event.is_set():
            try:
                msg = self._bus.recv(timeout=1.0)
            except TimeoutError:
                continue
            if msg is None or msg.is_remote_frame:
                continue
            with self._recv_lock:
                self._notify(msg)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script for robodyno."""

import os
from setuptools import setup

from src.robodyno import __version__

with open(
    os.path.join(os.path.dirname(__file__), 'README.md'),
    mode='r',
    encoding='utf-8',
) as f:
    long_description = f.read()

setup(
    name='robodyno',
    version=__version__,
    maintainer='robottime',
    maintainer_email='lab@robottime.cn',
    author='song',
    author_email='zhaosongy@126.com',
    description='The Robodyno Robot SDK for Python 3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://101.42.250.169/',
    keywords=['robodyno', 'robot', 'robot module'],
    license='Apache License, Version 2.0',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Framework :: Robot Framework',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    package_dir={
        '': 'src',
    },
    packages=[
        'robodyno',
        'robodyno.components',
        'robodyno.components.config',
        'robodyno.components.can_bus',
        'robodyno.components.webots',
        'robodyno.interfaces',
        'robodyno.tools',
        'robodyno.robots',
        'robodyno.robots.four_dof_palletizing_robot',
        'robodyno.robots.four_dof_scara_robot',
        'robodyno.robots.six_dof_collaborative_robot',
        'robodyno.robots.three_dof_delta_robot',
        'robodyno.robots.three_dof_palletizing_robot',
        'robodyno.robots.three_dof_cartesian_robot',
        'robodyno.robots.utils',
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.10.0',
        'rich >= 12.6.0',
        'click >= 7.1.2',
        'python-can>=3.2.0, <4.0',
        'importlib-metadata',
    ],
    extras_require={
        ':sys_platform == "win32"': ['candle-bus'],
    },
    entry_points={
        'robodyno.components.can_bus': [
            'Motor = robodyno.components.can_bus.motor:Motor',
            'PwmDriver = robodyno.components.can_bus.pwm_driver:PwmDriver',
            'StepperDriver = robodyno.components.can_bus.stepper_driver:StepperDriver',
            'SliderModule = robodyno.components.can_bus.slider_module:SliderModule',
            'ImuSensor = robodyno.components.can_bus.imu_sensor:ImuSensor',
            'GpsSensor = robodyno.components.can_bus.gps_sensor:GpsSensor',
            'LedDriver = robodyno.components.can_bus.led_driver:LedDriver',
            'Battery = robodyno.components.can_bus.battery:Battery',
        ],
        'robodyno.components.webots': [
            'Motor = robodyno.components.webots.motor:Motor',
            'SliderModule = robodyno.components.webots.slider_module:SliderModule',
        ],
        'robodyno.robots': [
            'FourDoFPallet = robodyno.robots.four_dof_palletizing_robot.four_dof_pallet_robot:FourDoFPallet',
            'FourDoFScara = robodyno.robots.four_dof_scara_robot.four_dof_scara_robot:FourDoFScara',
            'SixDoFCollabRobot = robodyno.robots.six_dof_collaborative_robot.six_dof_collab_robot:SixDoFCollabRobot',
            'ThreeDoFDelta = robodyno.robots.three_dof_delta_robot.three_dof_delta_robot:ThreeDoFDelta',
            'ThreeDoFPallet = robodyno.robots.three_dof_palletizing_robot.three_dof_pallet_robot:ThreeDoFPallet',
            'ThreeDoFCartesian = robodyno.robots.three_dof_cartesian_robot.three_dof_cartesian_robot:ThreeDoFCartesian',
        ],
        'console_scripts': [
            'robodyno = robodyno.tools.cli:robodyno',
        ],
    },
)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides a class for controlling Robodyno motors.

The Motor class provided by this module is used to control Robodyno motors
through the CAN bus. It provides methods for setting motor parameters, reading motor
states, and controlling the motor to run in different modes.

Examples:

    >>> from robodyno.components import Motor
    >>> from robodyno.interfaces import CanBus
    >>> can_bus = CanBus()
    >>> motor = Motor(can_bus)
    >>> print(motor)
    Motor(id=0x10, type=ROBODYNO_PRO_P44, state=DISABLED, error=NONE)
    >>> can_bus.disconnect()
"""

from time import sleep
from math import pi, fabs
from enum import Enum
from struct import unpack
from typing import Optional

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model
from robodyno.components.config.robottime_config import ROBOTTIME_PARAMS


class Motor(CanBusDevice):
    """Controls Robodyno motors through the CAN bus.

    Attributes:
        id (int): The ID of the motor device on the CAN bus.
        type (Model): The model type of the motor.
        fw_ver (float): The firmware version of the motor.
        reduction (float): The reduction ratio of the motor.
        available_velocity (float): The maximum velocity of the motor.
        available_torque (float): The maximum torque of the motor.
        available_current (float): The maximum phase current of the motor.
        torque_constant (float):
            The torque constant of the motor in Nm/A. This is the torque provided by
            the motor when the current is 1A.
        with_brake (bool): Whether the motor has a brake.
        state (Motor.MotorState): The working state of the motor.
        error (Dict): The error information of the motor.
        mode (Motor.MotorControlMode): The control mode of the motor.
        sync_time (int): The synchronization time of the motor in milliseconds.
    """

    _CMD_HEARTBEAT = 0x02
    _CMD_ESTOP = 0x03
    _CMD_REBOOT = 0x04
    _CMD_CLEAR_ERRORS = 0x05
    _CMD_SAVE = 0x06
    _CMD_CONFIG_CAN = 0x07
    _CMD_SET_STATE = 0x08
    _CMD_GET_HARDWARE_STATUS = 0x09
    _CMD_GET_MOTOR_FEEDBACK = 0x0A
    _CMD_GET_MODE = 0x0B
    _CMD_SET_MODE = 0x0C
    _CMD_GET_PID = 0x0D
    _CMD_SET_PID = 0x0E
    _CMD_GET_LIMITS = 0x0F
    _CMD_SET_LIMITS = 0x10
    _CMD_SET_POS = 0x11
    _CMD_SET_VEL = 0x12
    _CMD_SET_TORQUE = 0x13
    _CMD_RESET = 0x14
    _CMD_UNLOCK = 0x15
    _CMD_GET_ABS_POS = 0x16
    _CMD_SET_ABS_POS = 0x17
    _CMD_INIT_POS = 0x18
    _CMD_INIT_ABS_POS = 0x19
    _CMD_GET_STATE = 0x1A

    def __init__(self, can: CanBus, id_: int = 0x10, type_: Optional[str] = None):
        """Initializes the Robodyno motor based on the device id and type.

        Args:
            can (CanBus): The CAN bus interface.
            id_ (int): The device id. The default factory id of the motor is 0x10.
            type_ (string): Forcing the motor type. The default is None.

        Raises:
            ValueError: If the motor type is invalid.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if type_:
            self.type = getattr(Model, type_.upper(), None)
        if self.type not in ROBOTTIME_PARAMS['motor']:
            valid_models = ', '.join(
                map(lambda model: model.name, ROBOTTIME_PARAMS['motor'].keys())
            )
            raise ValueError(
                f'Invalid motor type. Valid motor types are: {valid_models}.'
            )

        self.__dict__.update(ROBOTTIME_PARAMS['motor'].get(self.type, None))

        self.state = self.MotorState.UNKNOWN
        self.error = {
            'error': self.MotorError.NONE,
            'motor_err': self.MotorMotorError.NONE,
            'encoder_err': self.MotorEncoderError.NONE,
            'controller_err': self.MotorControllerError.NONE,
        }
        self.mode = self.MotorControlMode.UNKNOWN
        self.sync_time = 0
        if self.fw_ver >= 1:
            self.get_state(1.5)

        self._rot_factor = -self.reduction / 2.0 / pi
        if self.fw_ver <= 0.3:
            self._rot_factor = -self._rot_factor
        self._can.subscribe(
            self._heartbeat_callback,
            device_id=self.id,
            cmd_id=self._CMD_HEARTBEAT,
        )

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}'
            f'(id=0x{self.id:02X}, type={self.type.name}, '
            f'state={self.state.name}, error={self.error["error"].name})'
        )

    def __del__(self):
        try:
            self._can.unsubscribe(
                self._heartbeat_callback,
                device_id=self.id,
                cmd_id=self._CMD_HEARTBEAT,
            )
        except KeyError:
            pass

    def _heartbeat_callback(self, data, timestamp, device_id, command_id) -> None:
        """Updates the motor attributes when receiving the heartbeat command."""
        del device_id, command_id
        state, err, merr, eerr, cerr, cm, im, _ = unpack('<BBBBBBBB', data)
        self.state = self.MotorState(state)
        if self.fw_ver < 1:
            self.error.update(
                {
                    'error': self.MotorError(err + 1 if err > 0 else 0),
                    'motor_err': self.MotorMotorError(merr + 1 if merr > 0 else 0),
                    'encoder_err': self.MotorEncoderError(eerr + 1 if eerr > 0 else 0),
                    'controller_err': self.MotorControllerError(
                        cerr + 1 if cerr > 0 else 0
                    ),
                }
            )
        else:
            self.error.update(
                {
                    'error': self.MotorError(err),
                    'motor_err': self.MotorMotorError(merr),
                    'encoder_err': self.MotorEncoderError(eerr),
                    'controller_err': self.MotorControllerError(cerr),
                }
            )
        self.mode = self.MotorControlMode((cm, im))
        self.sync_time = timestamp

    def _set_state(self, state) -> None:
        """Sets the motor working state.

        Args:
            state (Motor.MotorState): The motor working state to set.
        """
        self._can.send(self.id, self._CMD_SET_STATE, 'I', state.value)

    def enable(self) -> None:
        """Enables the motor.

        After enabled, the motor can be controlled.
        """
        self._set_state(self.MotorState.ENABLED)
        sleep(0.02)

    def disable(self) -> None:
        """Disables the motor.

        This is the default state after power on. After disabled, the motor
        will stop working and loose torque.
        """
        self._set_state(self.MotorState.DISABLED)

    def unlock(self) -> None:
        """Unlocks the motor.

        This command only works on motors with brake. After unlocked, the motor
        can be rotated freely.

        Raises:
            NotImplementedError: If the motor does not support brake.
        """
        if not self.with_brake:
            raise NotImplementedError('unlock is not supported on motors without brake')
        self._can.send(self.id, self._CMD_UNLOCK, '')

    def init_pos(self, initial_pos: float = 0) -> None:
        """Sets the current position of the motor as the initial position.

        Args:
            initial_pos (float): The initial position in rad.

        Raises:
            NotImplementedError: If the firmware version is < 1.0.
        """
        if self.fw_ver < 1.0:
            raise NotImplementedError(
                'Initial position is not supported on firmware < 1.0'
            )
        self._can.send(self.id, self._CMD_INIT_POS, 'f', initial_pos * self._rot_factor)

    def init_abs_pos(self, initial_pos: float = 0) -> None:
        """Sets the current absolute position of the motor as the initial position.

        Args:
            initial_pos (float): The initial absolute position in rad.

        Raises:
            NotImplementedError: If the firmware version is < 1.0.
        """
        if self.fw_ver < 1.0:
            raise NotImplementedError(
                'Absolute position is not supported on firmware < 1.0'
            )
        self._can.send(
            self.id, self._CMD_INIT_ABS_POS, 'f', initial_pos * self._rot_factor
        )

    def calibrate(self) -> None:
        """Calibrates the motor.

        Save the configurations after calibration manually by calling
        `save_configuration()` if needed.
        """
        self._set_state(self.MotorState.CALIBRATE)

    def config_can_bus(
        self, new_id: int = None, heartbeat: int = 1000, bitrate: int = 1000000
    ) -> None:
        """Configures the CAN bus settings of the motor.

        Args:
            new_id (int): The new device id.
            heartbeat (int): The heartbeat period in milliseconds.
            bitrate (int): The bitrate of the CAN bus. Choose from 250000,
                500000, 1000000.

        Raises:
            ValueError: If the new CAN id is not in the range of 0x01-0x3f.
            ValueError: If the bitrate is not one of 250000, 500000, 1000000.
        """
        if new_id is None:
            new_id = self.id
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')

        bitrate_id = {
            250000: 0,
            500000: 1,
            1000000: 2,
        }
        if bitrate not in bitrate_id:
            raise ValueError('Bitrate must be one of 250000, 500000, 1000000.')
        self._can.send(
            self.id,
            self._CMD_CONFIG_CAN,
            'HHI',
            new_id,
            bitrate_id[bitrate],
            int(heartbeat),
        )

    def save(self) -> None:
        """Saves the motor configurations.

        If the motor is in a running state, the motor will stop and save the
        configurations.
        """
        self._can.send(self.id, self._CMD_SAVE, '')
        sleep(0.1)
        self.clear_errors()

    def save_configuration(self) -> None:
        """Saves the motor configurations.

        This method will be deprecated. Use `save` instead.
        """
        self.save()

    def clear_errors(self) -> None:
        """Clears the motor errors.

        Most errors are cleared automatically if the cause is resolved.
        """
        self._can.send(self.id, self._CMD_CLEAR_ERRORS, '')

    def estop(self) -> None:
        """Requests an emergency stop.

        The motor will loose torque immediately and report an ESTOP error.
        """
        self._can.send(self.id, self._CMD_ESTOP, '')

    def reboot(self) -> None:
        """Reboots the motor."""
        self._can.send(self.id, self._CMD_REBOOT, '')

    def reset(self) -> None:
        """Resets the motor.

        This command will reset the motor to the factory settings. All
        configurations will be lost and the id will be reset to 0x10.
        The motor must be calibrated again after reset.
        """
        self._can.send(self.id, self._CMD_RESET, '')

    def position_mode(self) -> None:
        """Sets the motor to position mode."""
        self._can.send(
            self.id,
            self._CMD_SET_MODE,
            'BB',
            *self.MotorControlMode.POSITION_MODE.value,
        )

    def position_filter_mode(self, bandwidth: float) -> None:
        """Sets the motor to position filter mode.

        Args:
            bandwidth (float): The bandwidth of the filter in Hz. The value is suggested
                to be the control loop frequency in Hz.
        """
        cmode, imode = self.MotorControlMode.POSITION_FILTER_MODE.value
        self._can.send(self.id, self._CMD_SET_MODE, 'BBf', cmode, imode, bandwidth)

    def position_track_mode(self, vel: float, acc: float, dec: float) -> None:
        """Sets the motor to position track mode.

        Args:
            vel (float): The velocity limit in rad/s.
            acc (float): The acceleration in rad/s^2. The value must be positive.
            dec (float): The deceleration in rad/s^2. The value must be positive.
        """
        cmode, imode = self.MotorControlMode.POSITION_TRACK_MODE.value
        self._can.send(
            self.id,
            self._CMD_SET_MODE,
            'BBeee',
            cmode,
            imode,
            fabs(vel * self._rot_factor),
            fabs(acc * self._rot_factor),
            fabs(dec * self._rot_factor),
        )

    def velocity_mode(self) -> None:
        """Sets the motor to velocity mode."""
        self._can.send(
            self.id,
            self._CMD_SET_MODE,
            'BB',
            *self.MotorControlMode.VELOCITY_MODE.value,
        )

    def velocity_ramp_mode(self, ramp: float) -> None:
        """Sets the motor to velocity ramp mode.

        Args:
            ramp (float): The velocity ramp in rad/s^2. The value must be positive.
        """
        cmode, imode = self.MotorControlMode.VELOCITY_RAMP_MODE.value
        self._can.send(
            self.id,
            self._CMD_SET_MODE,
            'BBf',
            cmode,
            imode,
            fabs(ramp * self._rot_factor),
        )

    def torque_mode(self) -> None:
        """Sets the motor to torque mode."""
        self._can.send(
            self.id, self._CMD_SET_MODE, 'BB', *self.MotorControlMode.TORQUE_MODE.value
        )

    def set_pos(self, pos: float, vel_ff: float = 0, torque_ff: float = 0) -> None:
        """Sets the target position of the motor.

        Args:
            pos (float): The target position in rad.
            vel_ff (float): The velocity feedforward in rad/s.
            torque_ff (float): The torque feedforward in Nm.
        """
        self._can.send(
            self.id,
            self._CMD_SET_POS,
            'fee',
            pos * self._rot_factor,
            vel_ff * self._rot_factor,
            torque_ff / self.reduction,
        )

    def set_abs_pos(self, pos: float, vel_ff: float = 0, torque_ff: float = 0) -> None:
        """Sets the target absolute position of the motor.

        Args:
            pos (float): The target absolute position in rad.
            vel_ff (float): The velocity feedforward in rad/s.
            torque_ff (float): The torque feedforward in Nm.

        Raises:
            NotImplementedError: If the firmware version is < 1.0.
        """
        if self.fw_ver < 1.0:
            raise NotImplementedError(
                'Absolute position is not supported on firmware < 1.0'
            )
        self._can.send(
            self.id,
            self._CMD_SET_ABS_POS,
            'fee',
            pos * self._rot_factor,
            vel_ff * self._rot_factor,
            torque_ff / self.reduction,
        )

    def set_vel(self, vel: float, torque_ff: float = 0) -> None:
        """Sets the target velocity of the motor.

        Args:
            vel (float): The target velocity in rad/s.
            torque_ff (float): The torque feedforward in Nm.
        """
        self._can.send(
            self.id,
            self._CMD_SET_VEL,
            'ff',
            vel * self._rot_factor,
            torque_ff / self.reduction,
        )

    def set_torque(self, torque: float) -> None:
        """Sets the target torque of the motor.

        Args:
            torque (float): The target torque in Nm.
        """
        self._can.send(self.id, self._CMD_SET_TORQUE, 'f', torque / self.reduction)

    def set_pid(self, pos_kp: float, vel_kp: float, vel_ki: float) -> None:
        """Sets the PID parameters of the motor.

        Args:
            pos_kp (float): The Kp of the position control loop.
            vel_kp (float): The Kp of the velocity control loop.
            vel_ki (float): The Ki of the velocity control loop.
        """
        self._can.send(self.id, self._CMD_SET_PID, 'fee', pos_kp, vel_kp, vel_ki)

    def set_vel_limit(self, vel_lim: float) -> None:
        """Sets the velocity limit of the motor.

        Args:
            vel_lim (float): The velocity limit in rad/s. The value must be positive.

        Raises:
            ValueError: If the velocity limit is out of range.
        """
        if not 0 < vel_lim <= self.available_velocity:
            raise ValueError(f'Velocity limit {vel_lim} is out of range.')
        self._can.send(
            self.id, self._CMD_SET_LIMITS, 'ff', fabs(vel_lim * self._rot_factor), 0
        )

    def set_current_limit(self, current_lim: float) -> None:
        """Sets the current limit of the motor.

        Args:
            current_lim (float): The current limit in amps. The value must be positive.

        Raises:
            ValueError: If the current limit is out of range.
        """
        if not 0 < current_lim <= self.available_current:
            raise ValueError(f'Current limit {current_lim} is out of range.')
        self._can.send(self.id, self._CMD_SET_LIMITS, 'ff', 0, current_lim)

    def get_state(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the motor state and error.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (tuple | None): The motor state, error, and control mode. Returns None if
                the read times out.

        Raises:
            NotImplementedError: If the firmware version is < 1.0.
        """
        if self.fw_ver < 1.0:
            raise NotImplementedError('get_state is not supported on firmware < 1.0')
        try:
            state, err, merr, eerr, cerr, cm, im, _ = self._can.get(
                self.id, self._CMD_GET_STATE, 'B' * 8, timeout
            )
        except TimeoutError:
            return None
        self.state = self.MotorState(state)
        self.error.update(
            {
                'error': self.MotorError(err),
                'motor_err': self.MotorMotorError(merr),
                'encoder_err': self.MotorEncoderError(eerr),
                'controller_err': self.MotorControllerError(cerr),
            }
        )
        self.mode = self.MotorControlMode((cm, im))
        return (self.state, self.error, self.mode)

    def get_voltage(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the DC bus voltage of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The DC bus voltage in volts. Returns None if the read
                times out.
        """
        try:
            vbus, _ = self._can.get(
                self.id, self._CMD_GET_HARDWARE_STATUS, 'ff', timeout
            )
        except TimeoutError:
            return None
        return vbus

    def get_temperature(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the temperature of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The temperature in degrees Celsius. Returns None if the
                read times out.
        """
        try:
            _, temperature = self._can.get(
                self.id, self._CMD_GET_HARDWARE_STATUS, 'ff', timeout
            )
        except TimeoutError:
            return None
        return temperature

    def get_mode(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the control mode and the mode parameters of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (tuple | None): The control mode and the mode parameters. Returns None if
                the read times out.
        """
        try:
            payload = self._can.get(self.id, self._CMD_GET_MODE, '8s', timeout)[0]
        except TimeoutError:
            return None
        control_mode, input_mode = unpack('<BB', payload[:2])
        mode = self.MotorControlMode((control_mode, input_mode))
        self.mode = mode
        if mode == self.MotorControlMode.POSITION_FILTER_MODE:
            (bandwidth,) = unpack('<f', payload[2:6])
            return (mode, {'bandwidth': bandwidth})
        elif mode == self.MotorControlMode.POSITION_TRACK_MODE:
            vel, acc, dec = unpack('<eee', payload[2:8])
            return (
                mode,
                {
                    'vel': fabs(vel / self._rot_factor),
                    'acc': fabs(acc / self._rot_factor),
                    'dec': fabs(dec / self._rot_factor),
                },
            )
        elif mode == self.MotorControlMode.VELOCITY_RAMP_MODE:
            (ramp,) = unpack('<f', payload[2:6])
            return (mode, {'ramp': fabs(ramp / self._rot_factor)})
        else:
            return (mode,)

    def get_feedback(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the feedbacks of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (tuple | None): The position, velocity, and torque. Returns None if the
                read times out.
        """
        try:
            pos, vel, torque = self._can.get(
                self.id, self._CMD_GET_MOTOR_FEEDBACK, 'fee', timeout
            )
        except TimeoutError:
            return None
        return pos / self._rot_factor, vel / self._rot_factor, torque * self.reduction

    def get_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the position of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The position in rad. Returns None if the read times
                out.
        """
        feedback = self.get_feedback(timeout)
        if feedback is None:
            return None
        return feedback[0]

    def get_abs_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the absolute position of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The absolute position in rad. Returns None if the
                read times out.

        Raises:
            NotImplementedError: If the firmware version is < 1.0.
        """
        if self.fw_ver < 1.0:
            raise NotImplementedError(
                'Absolute position is not supported on firmware < 1.0'
            )
        try:
            pos = self._can.get(self.id, self._CMD_GET_ABS_POS, 'f', timeout)[0]
        except TimeoutError:
            return None
        return pos / self._rot_factor

    def get_vel(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the velocity of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The velocity in rad/s. Returns None if the read times
                out.
        """
        feedback = self.get_feedback(timeout)
        if feedback is None:
            return None
        return feedback[1]

    def get_torque(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the torque of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The torque in Nm. Returns None if the read times out.
        """
        feedback = self.get_feedback(timeout)
        if feedback is None:
            return None
        return feedback[2]

    def get_pid(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the PID parameters of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (tuple | None): The position KP, velocity KP, and velocity KI. Returns
                None if the read times out.
        """
        try:
            pos_kp, vel_kp, vel_ki = self._can.get(
                self.id, self._CMD_GET_PID, 'fee', timeout
            )
        except TimeoutError:
            return None
        return (pos_kp, vel_kp, vel_ki)

    def get_vel_limit(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the velocity limit of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The velocity limit in rad/s. Returns None if the read
                times out.
        """
        try:
            vel_limit, _ = self._can.get(self.id, self._CMD_GET_LIMITS, 'ff', timeout)
        except TimeoutError:
            return None
        return vel_limit / fabs(self._rot_factor)

    def get_current_limit(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the current limit of the motor.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): The current limit in A. Returns None if the read
                times out.
        """
        try:
            _, current_limit = self._can.get(
                self.id, self._CMD_GET_LIMITS, 'ff', timeout
            )
        except TimeoutError:
            return None
        return current_limit

    class MotorState(Enum):
        """Enumerates motor working states."""

        UNKNOWN = -1
        """The state of the motor is unknown."""
        BOOTING = 0
        """The motor is booting."""
        DISABLED = 1
        """The motor is disabled."""
        CALIBRATE = 3
        """The motor is preparing for calibration."""
        MOTOR_CALIBRATING = 4
        """The motor is calibrating."""
        OFFSET_CALIBRATING = 7
        """The motor is calibrating the encoder."""
        ENABLED = 8
        """The motor is enabled."""
        HOMING = 11
        """The motor is aligning the encoders."""

    class MotorError(Enum):
        """Enumerates motor errors."""

        NONE = 0
        """No error."""
        INVALID_STATE = 1
        """The motor is in an invalid state."""
        UNDER_VOLTAGE = 2
        """The motor is under voltage."""
        OVER_VOLTAGE = 3
        """The motor is over voltage."""
        CURRENT_MEASUREMENT_TIMEOUT = 4
        """The motor current measurement timed out."""
        BRAKE_RESISTOR_DISARMED = 5
        """The brake resistor is disarmed."""
        MOTOR_DISARMED = 6
        """The motor is disarmed."""
        MOTOR_FAILED = 7
        """The driver of the motor failed."""
        SENSORLESS_ESTIMATOR_FAILED = 8
        """The sensorless estimator of the motor failed."""
        ENCODER_FAILED = 9
        """The encoder of the motor failed."""
        CONTROLLER_FAILED = 10
        """The controller of the motor failed."""
        POS_CTRL_DURING_SENSORLESS = 11
        """The position controller of the motor is running during sensorless mode."""
        WATCHDOG_TIMER_EXPIRED = 12
        """The watchdog timer of the motor expired."""
        MIN_ENDSTOP_PRESSED = 13
        """The minimum endstop of the motor is pressed."""
        MAX_ENDSTOP_PRESSED = 14
        """The maximum endstop of the motor is pressed."""
        ESTOP_REQUESTED = 15
        """The motor is requested to emergency stop."""
        HOMING_WITHOUT_ENDSTOP = 16
        """The motor is homing without endstop."""
        OVER_TEMP = 17
        """The motor is over temperature."""

    class MotorMotorError(Enum):
        """Enumerates motor driver errors."""

        NONE = 0
        """No error."""
        PHASE_RESISTANCE_OUT_OF_RANGE = 1
        """The phase resistance of the motor is out of range."""
        PHASE_INDUCTANCE_OUT_OF_RANGE = 2
        """The phase inductance of the motor is out of range."""
        ADC_FAILED = 3
        """The ADC of the motor failed."""
        DRV_FAULT = 4
        """The driver of the motor failed."""
        CONTROL_DEADLINE_MISSED = 5
        """The control deadline of the motor is missed."""
        NOT_IMPLEMENTED_MOTOR_TYPE = 6
        """The motor type is not implemented."""
        BRAKE_CURRENT_OUT_OF_RANGE = 7
        """The brake current of the motor is out of range."""
        MODULATION_MAGNITUDE = 8
        """The low level modulation failed."""
        BRAKE_DEADTIME_VIOLATION = 9
        """The brake deadtime of the motor is violated."""
        UNEXPECTED_TIMER_CALLBACK = 10
        """The timer callback of the motor is unexpected."""
        CURRENT_SENSE_SATURATION = 11
        """The current sense of the motor is saturated."""
        CURRENT_LIMIT_VIOLATION = 13
        """The current limit of the motor is violated."""
        BRAKE_DUTY_CYCLE_NAN = 14
        """The brake duty cycle of the motor is NaN."""
        DC_BUS_OVER_REGEN_CURRENT = 15
        """The DC bus is over allowable regenerative current."""
        DC_BUS_OVER_CURRENT = 16
        """The DC bus is over current."""

    class MotorEncoderError(Enum):
        """Enumerates motor encoder errors."""

        NONE = 0
        """No error."""
        UNSTABLE_GAIN = 1
        """The PLL gain of the encoder is unstable."""
        CPR_POLEPAIRS_MISMATCH = 2
        """The CPR and pole pairs of the encoder mismatch."""
        NO_RESPONSE = 3
        """The encoder does not respond."""
        UNSUPPORTED_ENCODER_MODE = 4
        """The encoder mode is not supported."""
        ILLEGAL_HALL_STATE = 5
        """The hall state of the encoder is illegal."""
        INDEX_NOT_FOUND_YET = 6
        """The index of the encoder is not found yet."""
        ABS_SPI_TIMEOUT = 7
        """The SPI of the encoder timed out."""
        ABS_SPI_COM_FAIL = 8
        """The SPI of the encoder failed."""
        ABS_SPI_NOT_READY = 9
        """The SPI of the encoder is not ready."""

    class MotorControllerError(Enum):
        """Enumerates motor controller errors."""

        NONE = 0
        """No error."""
        OVERSPEED = 1
        """The motor is overspeed."""
        INVALID_INPUT_MODE = 2
        """The input mode of the motor is invalid."""
        UNSTABLE_GAIN = 3
        """The gain of the sensorless estimator is unstable."""
        INVALID_MIRROR_AXIS = 4
        """The mirror axis of the motor is invalid."""
        INVALID_LOAD_ENCODER = 5
        """The encoder of the motor is invalid."""
        INVALID_ESTIMATE = 6
        """The estimate feedbacks of the motor is invalid."""

    class MotorControlMode(Enum):
        """Enumerates motor control modes."""

        UNKNOWN = (0, 0)
        """The motor control mode is unknown."""
        POSITION_MODE = (3, 1)
        """Control the motor in position mode."""
        POSITION_FILTER_MODE = (3, 3)
        """Control the motor in position filter mode."""
        POSITION_TRACK_MODE = (3, 5)
        """Control the motor in position track mode."""
        VELOCITY_MODE = (2, 1)
        """Control the motor in velocity mode."""
        VELOCITY_RAMP_MODE = (2, 2)
        """Control the motor in velocity ramp mode."""
        TORQUE_MODE = (1, 1)
        """Control the motor in torque mode."""
